/* Copyright(C) 2004 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "senna_in.h"

#include <stdio.h>

#define __USE_GNU /* O_DIRECT */
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>

#include "str.h"
#include "io.h"
#include "cache.h"

#define SEN_IO_IDSTR "SENNA:IO:00.001"

/* VA hack */
/* max aio request (/proc/sys/fs/aio-max-nr) */
#define MAX_REQUEST     (64*1024)

#define MEM_ALIGN    sen_cache_block

/* end VA hack */

static unsigned int pagesize = 0;

typedef struct _sen_io_fileinfo {
#ifdef WIN32
  HANDLE fh;
  HANDLE fmo;
  sen_mutex mutex;
#else /* WIN32 */
  int fd;
  dev_t dev;
  ino_t inode;
#endif /* WIN32 */
} fileinfo;

typedef struct _sen_io_header {
  char idstr[16];
  uint32_t header_size;
  uint32_t segment_size;
  uint32_t max_segment;
} io_header;

inline static unsigned int sen_get_pagesize(void);
inline static sen_rc sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize);
inline static void sen_fileinfo_init(fileinfo *fis, int nfis);
inline static int sen_opened(fileinfo *fi);
inline static sen_rc sen_close(fileinfo *fi);
inline static void * sen_mmap(fileinfo *fi, off_t offset, size_t length);
inline static int sen_msync(void *start, size_t length);
inline static int sen_munmap(void *start, size_t length);
inline static sen_rc sen_pread(fileinfo *fi, void *buf, size_t count, off_t offset);
inline static sen_rc sen_pwrite(fileinfo *fi, void *buf, size_t count, off_t offset);

sen_io *
sen_io_create(const char *path, uint32_t header_size, uint32_t segment_size,
	      uint32_t max_segment, sen_io_mode mode, uint32_t cache_size)
{
  sen_io *io;
  fileinfo *fis;
  unsigned int b, max_nfiles;
  uint32_t bs;
  io_header *header;
  if (!path || !*path || (strlen(path) > PATH_MAX - 4)) { return NULL; }
  if (!pagesize) { pagesize = sen_get_pagesize(); }
  for (b = pagesize; b < sizeof(io_header) + header_size; b += pagesize);
  bs = (b + segment_size - 1) / segment_size;
  max_nfiles = (unsigned int)(
    ((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1) 
    / SEN_IO_FILE_SIZE);
  if (!(fis = SEN_MALLOC(sizeof(fileinfo) * max_nfiles))) { return NULL; }
  sen_fileinfo_init(fis, max_nfiles);
  if (!sen_open(fis, path, O_RDWR|O_CREAT|O_TRUNC, SEN_IO_FILE_SIZE)) {
    if ((header = sen_mmap(fis, 0, b))) {
      header->header_size = header_size;
      header->segment_size = segment_size;
      header->max_segment = max_segment;
      memcpy(header->idstr, SEN_IO_IDSTR, 16);
      sen_msync(header, b);
      if ((io = SEN_MALLOC(sizeof(sen_io)))) {
	if ((io->maps = SEN_MALLOC(sizeof(sen_io_mapinfo) * max_segment))) {
	  memset(io->maps, 0, sizeof(sen_io_mapinfo) * max_segment);
	  strncpy(io->path, path, PATH_MAX);
	  io->header = header;
	  io->user_header = ((byte *) header) + sizeof(io_header);
	  io->base = b;
	  io->base_seg = bs;
	  io->mode = mode;
	  io->cache_size = cache_size;
	  io->fis = fis;
	  io->nmaps = 0;
	  io->count = 0;
	  return io;
	}
	SEN_FREE(io);
      }
      sen_munmap(header, b);
    }
    sen_close(fis);
  }
  SEN_FREE(fis);
  return NULL;
}

sen_io *
sen_io_open(const char *path, sen_io_mode mode, uint32_t cache_size)
{
  sen_io *io;
  struct stat s;
  fileinfo *fis;
  io_header *header;
  unsigned int b, max_nfiles;
  uint32_t header_size = 0, segment_size = 0, max_segment = 0, bs;
  if (!path || !*path || (strlen(path) > PATH_MAX - 4)) { return NULL; }
  if (!pagesize) { pagesize = sen_get_pagesize(); }
  {
    io_header h;
    int fd = open(path, O_RDWR);
    if (fd == -1) { return NULL; }
    if (fstat(fd, &s) != -1 && s.st_size >= sizeof(io_header)) {
      if (read(fd, &h, sizeof(io_header)) == sizeof(io_header)) {
	/* if (!memcmp(header->idstr, SEN_IO_IDSTR, 16)) */
	{
	  header_size = h.header_size;
	  segment_size = h.segment_size;
	  max_segment = h.max_segment;
	}
      }
    }
    close(fd);
    if (!segment_size) { return NULL; }
  }
  for (b = pagesize; b < sizeof(io_header) + header_size; b += pagesize);
  bs = (b + segment_size - 1) / segment_size;
  max_nfiles = (unsigned int)(
    ((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1) 
    / SEN_IO_FILE_SIZE);
  if (!(fis = SEN_MALLOC(sizeof(fileinfo) * max_nfiles))) { return NULL; }
  sen_fileinfo_init(fis, max_nfiles);
  if (!sen_open(fis, path, O_RDWR, SEN_IO_FILE_SIZE)) {
    if ((header = sen_mmap(fis, 0, b))) {
      if ((io = SEN_MALLOC(sizeof(sen_io)))) {
	if ((io->maps = SEN_MALLOC(sizeof(sen_io_mapinfo) * max_segment))) {
	  memset(io->maps, 0, sizeof(sen_io_mapinfo) * max_segment);
	  strncpy(io->path, path, PATH_MAX);
	  io->header = header;
	  io->user_header = ((byte *) header) + sizeof(io_header);
	  io->base = b;
	  io->base_seg = bs;
	  io->mode = mode;
	  io->cache_size = cache_size;
	  io->fis = fis;
	  io->nmaps = 0;
	  io->count = 0;
	  return io;
	}
	SEN_FREE(io);
      }
      sen_munmap(header, b);
    }
    sen_close(fis);
  }
  SEN_FREE(fis);
  return NULL;
}

sen_rc
sen_io_close(sen_io *io)
{
  int i;
  sen_io_mapinfo *mi;
  fileinfo *fi;
  uint32_t bs = io->base_seg;
  uint32_t max_segment = io->header->max_segment;
  uint32_t segment_size = io->header->segment_size;
  unsigned int max_nfiles = (unsigned int)(
    ((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1) 
    / SEN_IO_FILE_SIZE);
  for (mi = io->maps, i = max_segment; i; mi++, i--) {
    if (mi->map) {
      /* if (atomic_read(mi->nref)) { return STILL_IN_USE ; } */
      sen_munmap(mi->map, segment_size);
    }
  }
  sen_munmap(io->header, io->base);
  for (fi = io->fis, i = max_nfiles; i; fi++, i--) { sen_close(fi); }
  SEN_FREE(io->fis);
  SEN_FREE(io->maps);
  SEN_FREE(io);
  return sen_success;
}

uint32_t
sen_io_base_seg(sen_io *io)
{
  return io->base_seg;
}

const char *
sen_io_path(sen_io *io)
{
  return io->path;
}

void *
sen_io_header(sen_io *io)
{
  return io->user_header;
}

inline static void
itoh(unsigned int i, char *p, unsigned int len)
{
  static const char *hex = "0123456789ABCDEF";
  p += len;
  *p-- = '\0';
  while (len--) {
    *p-- = hex[i & 0xf];
    i >>= 4;
  }
}

inline static void
gen_pathname(const char *path, char *buffer, int fno)
{
  size_t len = strlen(path);
  memcpy(buffer, path, len);
  if (fno) {
    buffer[len] = '.';
    itoh(fno, buffer + len + 1, 3);
  } else {
    buffer[len] = '\0';
  }
}

sen_rc
sen_io_size(sen_io *io, off_t *size)
{
  int fno;
  struct stat s;
  off_t tsize = 0;
  char buffer[PATH_MAX];
  uint32_t bs = io->base_seg;
  uint32_t max_segment = io->header->max_segment;
  uint32_t segment_size = io->header->segment_size;
  unsigned int max_nfiles = (unsigned int)(
    ((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1) 
    / SEN_IO_FILE_SIZE);
  for (fno = 0; fno < max_nfiles; fno++) {
    gen_pathname(io->path, buffer, fno);
    if (!stat(buffer, &s)) { tsize += s.st_size; }
  }
  *size = tsize;
  return sen_success;
}

sen_rc
sen_io_remove(const char *path)
{
  struct stat s;
  if (!stat(path, &s) && !unlink(path)) {
    int fno;
    char buffer[PATH_MAX];
    for (fno = 1; ; fno++) {
      gen_pathname(path, buffer, fno);
      if (stat(buffer, &s) || unlink(buffer)) { break; }
    }
    return sen_success;
  }
  return sen_file_operation_error;
}

sen_rc
sen_io_rename(const char *old_name, const char *new_name)
{
  struct stat s;
  if (!stat(old_name, &s) && !rename(old_name, new_name)) {
    int fno;
    char old_buffer[PATH_MAX];
    char new_buffer[PATH_MAX];
    for (fno = 1; ; fno++) {
      gen_pathname(old_name, old_buffer, fno);
      if (!stat(old_buffer, &s)) {
	gen_pathname(new_name, new_buffer, fno);
	rename(old_buffer, new_buffer);
      } else {
	break;
      }
    }
    return sen_success;
  }
  return sen_file_operation_error;
}

void *
sen_io_win_map(sen_io *io, sen_io_win *iw, uint32_t segment,
	       uint32_t offset, uint32_t size, sen_io_rw_mode mode)
{
  byte *p;
  off_t pos;
  int fno, base;
  uint32_t nseg, bseg;
  uint32_t segment_size = io->header->segment_size;
  uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
  iw->diff = 0;
  if (offset >= segment_size) {
    segment += offset / segment_size;
    offset = offset % segment_size;
  }
  nseg = (offset + size + segment_size - 1) / segment_size;
  bseg = segment + io->base_seg;
  fno = bseg / segments_per_file;
  base = fno ? 0 : io->base - io->base_seg * segment_size;
  pos = (bseg % segments_per_file) * segment_size + offset + base;
  if (!size || !io || segment + nseg > io->header->max_segment ||
      fno != (bseg + nseg - 1) / segments_per_file) {
    return NULL;
  }
  switch (mode) {
  case sen_io_rdonly:
    {
      fileinfo *fi = &io->fis[fno];
      if (!sen_opened(fi)) {
	char path[PATH_MAX];
	gen_pathname(io->path, path, fno);
	if (sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {
	  return NULL;
	}
      }
      if (!(p = SEN_MALLOC(size))) { return NULL; }
      if (sen_pread(fi, p, size, pos)) {
	SEN_FREE(p);
	return NULL;
      }
      iw->addr = p;
    }
    break;
  case sen_io_rdwr:
    // if (nseg > 1) { /* auto unmap is not implemented yet */
    if (nseg > 0) {
      fileinfo *fi = &io->fis[fno];
      if (!sen_opened(fi)) {
	char path[PATH_MAX];
	gen_pathname(io->path, path, fno);
	if (sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {
	  return NULL;
	}
      }
      if (!(p = sen_mmap(fi, pos, nseg * segment_size))) {
	return NULL;
      }
    } else {
      if (!(p = sen_io_seg_ref(io, segment))) { return NULL; }
    }
    iw->addr = p + offset;
    break;
  case sen_io_wronly: 
    if (!(p = SEN_MALLOC(size))) { return NULL; }
    memset(p, 0, size);
    iw->addr = p;
    break;
  default :
    return NULL;
  }
  iw->io = io;
  iw->mode = mode;
  iw->segment = segment;
  iw->offset = offset;
  iw->nseg = nseg;
  iw->size = size;
  iw->pos = pos;
  return iw->addr;
}

#ifdef USE_AIO
sen_rc
sen_io_win_mapv(sen_io_win **list, int nent)
{
  int i;
  sen_io_win *iw;
  struct aiocb *iocbs[MAX_REQUEST];
  struct aiocb iocb[MAX_REQUEST];
  CacheIOOper oper[MAX_REQUEST]; 
  int count = 0;

  sen_io_win **clist = list;
  int cl = 0;

retry:
  for (i = 0; i < nent; i++) {
    iw = list[i];
    if (sen_aio_enabled && iw->mode == sen_io_rdonly) {
        /* this block is same as sen_io_win_map() */
        sen_io *io = iw->io;        
        uint32_t segment = iw->segment, offset = iw->offset, size = iw->size;
        byte *p;
        off_t pos;
        int fno, base;
        uint32_t nseg, bseg;
        uint32_t segment_size = io->header->segment_size;
        uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
        fileinfo *fi;
        iw->diff = 0;
        if (offset >= segment_size) {
            segment += offset / segment_size;
            offset = offset % segment_size;
        }
	nseg = (offset + size + segment_size - 1) / segment_size;
	bseg = segment + io->base_seg;
	fno = bseg / segments_per_file;
	base = fno ? 0 : io->base - io->base_seg * segment_size;
	pos = (bseg % segments_per_file) * segment_size + offset + base;
        if (!size || !io || segment + nseg > io->header->max_segment ||
            fno != (bseg + nseg - 1) / segments_per_file) {
	  return sen_other_error;
        }
        fi = &io->fis[fno];
        if (!sen_opened(fi)) {
            char path[PATH_MAX];
	    gen_pathname(io->path, path, fno);
            if (sen_open(fi, path, O_RDWR|O_CREAT|O_DIRECT, SEN_IO_FILE_SIZE)) {
                return sen_other_error;
            }
        }
        {
            /* AIO + DIO + cache hack */
            /* calc alignment */
            off_t voffset = pos - (pos % MEM_ALIGN);
            uint32_t vsize = pos + size;

            vsize = ((vsize - 1) / MEM_ALIGN + 1) * MEM_ALIGN;
            vsize = vsize - voffset;

            /* diff of aligned offset */
            iw->diff = pos - voffset;

            dp ("pos: %lu, allocate: %d, really needed: %d\n", voffset, vsize, size);
            memset(&oper[count], 0, sizeof (CacheIOOper));
            memset(&iocb[count], 0, sizeof (struct aiocb));
            oper[count].iocb = &iocb[count];
            iocb[count].aio_fildes = fi->fd;
            iocb[count].aio_lio_opcode = LIO_READ;

            if (vsize <= MEM_ALIGN &&
                (p = sen_cache_read (&oper[count], fi->dev, fi->inode, voffset, vsize)) != NULL) {
                /* use cache process  */
                iw->cached = oper[count].cd->num;

                /* Cache require aio_read() or
                   already aio_read() by other process */
                if (oper[count].read == 1) {
                    iocbs[count] = &iocb[count];
                    count++; /* aio count */
                } else if (oper[count].cd->flag == CACHE_READ) {
                    /* this iw is ignored in this loop.
                       should re-check after AIO */
                    clist[cl++] = iw;
                }
            } else {
                /* This size cannot use Cache */
                dp ("Wont use cache offset=%lu size=%u\n", voffset, vsize);

                /* allocate aligned memory */
                if (posix_memalign(&p, MEM_ALIGN, vsize) != 0) {
                    return sen_other_error;
                }
                iocb[count].aio_buf = p;
                iocb[count].aio_nbytes = vsize;
                iocb[count].aio_offset = voffset;
                iocbs[count] = &iocb[count];

                /* This is not cached  */
                oper[count].cd = NULL;
                iw->cached = -1; 

                /* aio count up */
                count++;
            }
            iw->addr = p;
            iw->segment = segment;
            iw->offset = offset;
            iw->nseg = nseg;
            iw->size = size;
            iw->pos = pos;
        } /* End  AIO + DIO + cache hack */
    } else {
        if (!sen_io_win_map(iw->io, iw, iw->segment, iw->offset, iw->size, iw->mode)) {
            return sen_other_error;
        }
    }
  }

  if (sen_aio_enabled) {      
      if (count > 0) {
          int c;

          /* aio_read () */
          if (lio_listio (LIO_WAIT, iocbs, count, NULL) < 0) {
              perror ("lio_listio");
              return sen_other_error;
          }
          
          for (c=0;c<count;c++) {
              /* cache data is now VALID */
              if (oper[c].cd) oper[c].cd->flag = CACHE_VALID;
          }
      }
      if (cl > 0) {
          /*
           *  if previous loop have CACHE_READ CacheData,
           *  it should retry.
           */
          dp ("-- Validate Reading state CacheData (%d) --\n", cl);
          /* update list and nent for loop */
          list = clist; /* list of iw which use CACHE_READ CacheData */
          nent = cl;    /* number of iw */
          cl = 0; 
          count = 0;
          usleep(1);
          goto retry;
      } else
          dp("-- No Reading state CacheData. --\n");
  }
  return sen_success;
}
#endif /* USE_AIO */

sen_rc
sen_io_win_unmap(sen_io_win *iw)
{
  sen_rc rc = sen_success;
  sen_io *io = iw->io;
  uint32_t segment_size = io->header->segment_size;
  uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
  int nseg = iw->nseg;
  switch (iw->mode) {
  case sen_io_rdonly:
#ifdef USE_AIO
    /* VA hack */
    if (!sen_aio_enabled || (iw->cached < 0 && iw->addr)) { SEN_FREE(iw->addr); }
    else if (iw->cached >= 0){
      /* this data is cached */
      sen_cache_data_unref (iw->cached);
      iw->cached = -1;
    }
    /* end VA hack */
#else /* USE_AIO */
    if (iw->addr) { SEN_FREE(iw->addr); }
#endif /* USE_AIO */
    iw->addr = NULL;
    break;
  case sen_io_rdwr:
    // if (nseg > 1) { /* auto unmap is not implemented yet */
    if (nseg > 0) {
      sen_munmap(((byte *)iw->addr) - iw->offset, nseg * segment_size);
    } else {
      rc = sen_io_seg_unref(io, iw->segment);
#ifdef USE_AIO
      if (sen_aio_enabled) {
	int fno = (iw->segment + io->base_seg) / segments_per_file;
	fileinfo *fi = &io->fis[fno];
	sen_cache_mark_invalid(fi->dev, fi->inode, iw->pos, iw->size);
      }
#endif /* USE_AIO */
    }
    iw->addr = NULL;
    break;
  case sen_io_wronly: 
    {
      int fno = (iw->segment + io->base_seg) / segments_per_file;
      fileinfo *fi = &io->fis[fno];
      if (!sen_opened(fi)) {
	char path[PATH_MAX];
	gen_pathname(io->path, path, fno);
	rc = sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE);
      }
      if (!rc) {
	if (!(rc = sen_pwrite(fi, iw->addr, iw->size, iw->pos))) {
	  SEN_FREE(iw->addr);
	  iw->addr = NULL;
	}
#ifdef USE_AIO
        if (sen_aio_enabled) {
	  sen_cache_mark_invalid(fi->dev, fi->inode, iw->pos, iw->size);
	}
#endif /* USE_AIO */
      }
    }
    break;
  default :
    rc = sen_invalid_argument;
  }
  return rc;
}

#define SEG_MAP(io,segno,info)\
{\
  uint32_t segment_size = io->header->segment_size;\
  uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;\
  uint32_t bseg = segno + io->base_seg;\
  uint32_t fno = bseg / segments_per_file;\
  uint32_t base = fno ? 0 : io->base - io->base_seg * segment_size;\
  off_t pos = (bseg % segments_per_file) * segment_size + base;\
  fileinfo *fi = &io->fis[fno];\
  if (!sen_opened(fi)) {\
    char path[PATH_MAX];\
    gen_pathname(io->path, path, fno);\
    if (!sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {\
      if ((info->map = sen_mmap(fi, pos, segment_size))) {\
	uint32_t nmaps;\
	SEN_ATOMIC_ADD_EX(&io->nmaps, 1, nmaps);\
      }\
    }\
  } else {\
    if ((info->map = sen_mmap(fi, pos, segment_size))) {\
      uint32_t nmaps;\
      SEN_ATOMIC_ADD_EX(&io->nmaps, 1, nmaps);\
    }\
  }\
}

void
sen_io_seg_map_(sen_io *io, uint32_t segno, sen_io_mapinfo *info)
{
  SEG_MAP(io, segno, info);
}

void * 
sen_io_seg_ref(sen_io *io, uint32_t segno)
{
  uint32_t retry;
  sen_io_mapinfo *info;
  if (segno >= io->header->max_segment) { return NULL; }
  info = &io->maps[segno];
  for (retry = 0;; retry++) {
    uint32_t nref;
    SEN_ATOMIC_ADD_EX(&info->nref, 1, nref);
    if (nref >= SEN_IO_MAX_REF) {
      SEN_ATOMIC_ADD_EX(&info->nref, -1, nref);
      if (retry >= SEN_IO_MAX_RETRY) {
	sen_log("deadlock detected! in sen_io_seg_ref(%p, %u)", io, segno);
	return NULL;
      }
      usleep(1);
      continue;
    }
    if (!info->map) {
      if (nref) {
	SEN_ATOMIC_ADD_EX(&info->nref, -1, nref);
        if (retry >= SEN_IO_MAX_RETRY) {
	  sen_log("deadlock detected!! in sen_io_seg_ref(%p, %u)", io, segno);
	  return NULL;
	}
        usleep(1);
        continue;
      } else {
	SEG_MAP(io, segno, info);
      }
    }
    break;
  }
  info->count = io->count++;
  return info->map;
}

sen_rc
sen_io_seg_unref(sen_io *io, uint32_t segno)
{
  uint32_t nref;
  sen_io_mapinfo *info;
  if (segno >= io->header->max_segment) { return sen_invalid_argument; }
  info = &io->maps[segno];
  SEN_ATOMIC_ADD_EX(&info->nref, -1, nref);
  return sen_success;
}

sen_rc
sen_io_seg_expire(sen_io *io, uint32_t segno)
{
  uint32_t retry;
  sen_io_mapinfo *info;
  if (segno >= io->header->max_segment) { return sen_invalid_argument; }
  info = &io->maps[segno];
  for (retry = 0;; retry++) {
    uint32_t nref;
    SEN_ATOMIC_ADD_EX(&info->nref, 1, nref);
    if (nref) {
      SEN_ATOMIC_ADD_EX(&info->nref, -1, nref);
      if (retry >= SEN_IO_MAX_RETRY) {
	sen_log("deadlock detected! in sen_io_seg_expire(%p, %u)", io, segno);
	return sen_other_error;
      }
      usleep(1);
      continue;
    } else {
      SEN_ATOMIC_ADD_EX(&info->nref, SEN_IO_MAX_REF, nref);
      if (nref > 1) {
	SEN_ATOMIC_ADD_EX(&info->nref, -(SEN_IO_MAX_REF + 1), nref);
	if (retry >= SEN_IO_MAX_RETRY) {
	  sen_log("deadlock detected!! in sen_io_seg_expire(%p, %u)", io, segno);
	  return sen_other_error;
	}
        usleep(1);
        continue;
      } else {
	uint32_t nmaps;
	sen_munmap(info->map, io->header->segment_size);
	info->map = NULL;
	SEN_ATOMIC_ADD_EX(&info->nref, -(SEN_IO_MAX_REF + 1), nref);
	SEN_ATOMIC_ADD_EX(&io->nmaps, -1, nmaps);
        break;
      }
    }
  }
  return sen_success;
}

/** mmap abstraction **/

static size_t mmap_size = 0;

#ifdef WIN32

inline static unsigned int
sen_get_pagesize(void)
{
  SYSTEM_INFO si;
  GetSystemInfo( &si );
  return (unsigned int)si.dwAllocationGranularity;
}

inline static sen_rc
sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize)
{
  char *pd, *ps, buf[PATH_MAX];
  /* may be wrong if flags is just only O_RDWR */
  fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
		      FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
		      OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
  if (fi->fh == INVALID_HANDLE_VALUE) { return sen_file_operation_error; }
  if ((flags & O_TRUNC)) {
    CloseHandle(fi->fh);
    /* unable to assign OPEN_ALWAYS and TRUNCATE_EXISTING at once */
    fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
			TRUNCATE_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (fi->fh == INVALID_HANDLE_VALUE) { return sen_file_operation_error; }
  }
  for (pd = buf, ps = path; *ps && pd < buf + PATH_MAX - 1; ps++, pd++) {
    *pd = *ps == '\\' ? '_' : *ps;
  }
  *pd = '\0';
  /* signature may be wrong.. */
  fi->fmo = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, buf);
  /* open failed */
  if (fi->fmo == NULL) {
    // flock
    /* retry to open */
    fi->fmo = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, buf);
    /* failed again */
    if (fi->fmo == NULL) {
      /* try to create fmo */
      fi->fmo = CreateFileMapping(fi->fh, NULL, PAGE_READWRITE, 0, SEN_IO_FILE_SIZE, buf);
    }
    // funlock
  }
  if (fi->fmo != NULL) {
    if (GetLastError() != ERROR_ALREADY_EXISTS ) {
      MUTEX_INIT(fi->mutex);
      return sen_success;
    } else {
      sen_log("fmo object already exists! handle=%d", fi->fh);
      CloseHandle(fi->fmo);
    }
  } else {
    sen_log("failed to get FileMappingObject #%d", GetLastError());
  }
  CloseHandle(fi->fh);
  return sen_file_operation_error;
}

inline static void
sen_fileinfo_init(fileinfo *fis, int nfis)
{
  for (; nfis--; fis++) {
    fis->fh = INVALID_HANDLE_VALUE;
    fis->fmo = NULL;
  }
}

inline static int
sen_opened(fileinfo *fi)
{
  return fi->fh != INVALID_HANDLE_VALUE;
}

inline static sen_rc
sen_close(fileinfo *fi)
{
  if (fi->fmo != NULL) {
    CloseHandle(fi->fmo);
    fi->fmo = NULL;
  }
  if (fi->fh != INVALID_HANDLE_VALUE) {
    CloseHandle(fi->fh);
    MUTEX_DESTROY(fi->mutex);
    fi->fh = INVALID_HANDLE_VALUE;
  }
  return sen_success;
}

inline static void *
sen_mmap(fileinfo *fi, off_t offset, size_t length)
{
  void *res;
  /* file must be exceeded to SEN_IO_FILE_SIZE when FileMappingObject created.
     and, after fmo created, it's not allowed to expand the size of file.
  DWORD tail = (DWORD)(offset + length);
  DWORD filesize = GetFileSize(fi->fh, NULL);
  if (filesize < tail) {
    if (SetFilePointer(fi->fh, tail, NULL, FILE_BEGIN) != tail) {
      sen_log("SetFilePointer failed");
      return NULL;
    }
    if (!SetEndOfFile(fi->fh)) {
      sen_log("SetEndOfFile failed");
      return NULL;
    }
    filesize = tail;
  }
  */
  res = MapViewOfFile(fi->fmo, FILE_MAP_WRITE, 0, (DWORD)offset, (SIZE_T)length);
  if (res) {
    mmap_size += length;
  } else {
    sen_log("MapViewOfFile failed #%d", GetLastError());
  }
  return res;
}

inline static int
sen_msync(void *start, size_t length)
{
  /* return value may be wrong... */
  return FlushViewOfFile(start, length);
}

inline static int
sen_munmap(void *start, size_t length)
{
  if (UnmapViewOfFile(start)) {
    mmap_size -= length;
    return 0;
  } else {
    sen_log("munmap(%p,%d) failed <%d>", start, length, mmap_size);
    return -1;
  }
}

inline static sen_rc
sen_pread(fileinfo *fi, void *buf, size_t count, off_t offset)
{
  DWORD r, len;
  sen_rc rc = sen_success;
  MUTEX_LOCK(fi->mutex);
  r = SetFilePointer(fi->fh, offset, NULL, FILE_BEGIN);
  if (r == INVALID_SET_FILE_POINTER) {
    rc = sen_file_operation_error;
    sen_log("SetFilePointer error(%d)", GetLastError());
  } else {
    if (!ReadFile(fi->fh, buf, (DWORD)count, &len, NULL)) {
      rc = sen_file_operation_error;
      sen_log("ReadFile error(%d)", GetLastError());
    } else if (len != count) {
      sen_log("ReadFile %d != %d", count, len);
      rc = sen_file_operation_error;
    }
  }
  MUTEX_UNLOCK(fi->mutex);
  return rc;
}

inline static sen_rc
sen_pwrite(fileinfo *fi, void *buf, size_t count, off_t offset)
{
  DWORD r, len;
  sen_rc rc = sen_success;
  MUTEX_LOCK(fi->mutex);
  r = SetFilePointer(fi->fh, offset, NULL, FILE_BEGIN);
  if (r == INVALID_SET_FILE_POINTER) {
    rc = sen_file_operation_error;
    sen_log("SetFilePointer error(%d)", GetLastError());
  } else {
    if (!WriteFile(fi->fh, buf, (DWORD)count, &len, NULL)) {
      sen_log("WriteFile error(%d)", GetLastError());
      rc = sen_file_operation_error;
    } else if (len != count) {
      sen_log("WriteFile %d != %d", count, len);
      rc = sen_file_operation_error;
    }
  }
  MUTEX_UNLOCK(fi->mutex);
  return rc;
}

#else /* WIN32 */

inline static unsigned int
sen_get_pagesize(void)
{
  return (unsigned int)sysconf(_SC_PAGESIZE);
}

inline static sen_rc
sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize)
{
  struct stat st;    
  if ((fi->fd = open(path, flags, 0666)) == -1) {
    return sen_file_operation_error;
  }
  fstat(fi->fd, &st);
  fi->dev = st.st_dev;
  fi->inode = st.st_ino;
  return sen_success;
}

inline static void
sen_fileinfo_init(fileinfo *fis, int nfis)
{
  for (; nfis--; fis++) { fis->fd = -1; }
}

inline static int
sen_opened(fileinfo *fi)
{
  return fi->fd != -1;
}

inline static sen_rc
sen_close(fileinfo *fi)
{
  if (fi->fd != -1) {
    close(fi->fd);
    fi->fd = -1;
  }
  return sen_success;
}

inline static void *
sen_mmap(fileinfo *fi, off_t offset, size_t length)
{
  void *res;
  struct stat s;
  off_t tail = offset + length;
  if ((fstat(fi->fd, &s) == -1) ||
      (s.st_size < tail && ftruncate(fi->fd, tail) == -1)) {
    sen_log("fstat or ftruncate failed %d", fi->fd);
    return NULL;
  }
  res = mmap(NULL, length, PROT_READ|PROT_WRITE, MAP_SHARED, fi->fd, offset);
  if (MAP_FAILED == res) {
    sen_log("mmap(%d,%d,%d)=%p <%d>", length, fi->fd, offset, res, mmap_size);
    return NULL;
  }
  mmap_size += length;
  return res;
}

inline static int
sen_msync(void *start, size_t length)
{
  return msync(start, length, MS_SYNC);
}

inline static int
sen_munmap(void *start, size_t length)
{
  int res;
  res = munmap(start, length);
  if (res) {
    sen_log("munmap(%p,%d) failed <%d>", start, length, mmap_size);
  } else {
    mmap_size -= length;
  }
  return res;
}

inline static sen_rc
sen_pread(fileinfo *fi, void *buf, size_t count, off_t offset)
{
  if (pread(fi->fd, buf, count, offset) != count) {
    return sen_file_operation_error;
  }
  return sen_success;
}

inline static sen_rc
sen_pwrite(fileinfo *fi, void *buf, size_t count, off_t offset)
{
  if (pwrite(fi->fd, buf, count, offset) != count) {
    return sen_file_operation_error;
  }
  return sen_success;
}

#endif /* WIN32 */
